#!/usr/bin/perl -w
use strict;
use IO::Socket::INET;
use Event qw(loop);
use POSIX qw(EINTR);

package Connection;
use POSIX qw(EINTR EWOULDBLOCK);

sub new {
    my ($class, $socket, $out) = @_;
    $socket->blocking(0);
    my $connection = bless {
        handle	   => $socket,
        rd_watcher => undef,
        wr_watcher => undef,
        in_buffer  => "",
        out_buffer => $out,
        peer_ip    => $socket->peerhost,
        peer_port  => $socket->peerport,
    }, $class;
    $connection->{rd_watcher} = Event->io(fd => $connection->{handle},
                                          poll => "r",
                                          cb => [$connection, "readable"]);
    $connection->{wr_watcher} = Event->io(fd => $connection->{handle},
                                          poll => "w",
                                          cb => [$connection, "writable"]);
    print STDERR "Handling connection from $connection->{peer_ip}:$connection->{peer_port}\n";
    return $connection;
}

sub readable {
    my $connection = shift;

    my $rc = sysread($connection->{handle}, my $buffer, 8192);
    if ($rc) {
        $connection->{in_buffer} .= $buffer;
        return;
    }
    if (defined($rc)) {
        # EOF
        printf STDERR "Read %d bytes\n", length($connection->{in_buffer});
        $connection->{rd_watcher}->cancel;
        $connection->{rd_watcher} = undef;
        shutdown($connection->{handle}, 0);
        return;
    }
    return if $! == EINTR || $! == EWOULDBLOCK;
    print STDERR "Error reading from socket to $connection->{peer_ip}:$connection->{peer_port}: $!\n";
    $connection->drop;
}

sub writable {
    my $connection = shift;

    if (defined(my $rc = syswrite($connection->{handle},
                                  $connection->{out_buffer}))) {
        substr($connection->{out_buffer}, 0, $rc, "");
        if ($connection->{out_buffer} eq "") {
            $connection->{wr_watcher}->cancel;
            $connection->{wr_watcher} = undef;
            shutdown($connection->{handle}, 1);
        }
        return;
    }
    return if $! == EINTR || $! == EWOULDBLOCK;
    print STDERR "Error writing to socket to $connection->{peer_ip}:$connection->{peer_port}: $!\n";
    $connection->drop;
}

sub drop {
    my $connection = shift;
    if ($connection->{rd_watcher}) {
        $connection->{rd_watcher}->cancel;
        $connection->{rd_watcher} = undef;
    }
    if ($connection->{wr_watcher}) {
        $connection->{wr_watcher}->cancel;
        $connection->{wr_watcher} = undef;
    }
    close $connection->{handle};
}

sub DESTROY {
    my $connection = shift;
    print STDERR "Connection to $connection->{peer_ip}:$connection->{peer_port} forgotten\n";
}

package main;

$SIG{PIPE} = "IGNORE";

defined(my $file = shift) || die "Usage: $0 file\n";
my $out = do { 
    open(my $fh, "<", $file) || die "Could not open $file: $!";
    local $/; 
    <$fh>;
};

my $listener = IO::Socket::INET->new(LocalAddr => "127.0.0.1",
                                     LocalPort => 1234,
                                     Listen    => 5) ||
    die "Could not listen on 127.0.0.1:1234: $!\n";

$listener->blocking(0);
Event->io(fd => $listener,
          poll => "r",
          cb => sub {
              if (my $socket = $listener->accept) {
                  Connection->new($socket, $out);
              } else {
                  print STDERR "Error accepting connection: $!\n" unless
                      $! == EINTR;
              }
          });

loop;
