#!/usr/bin/perl -w
use strict;
use IO::Socket::INET;
use POSIX qw(EINTR);

my $rd_mask = my $wr_mask = "";
my %connections;

package Connection;
use POSIX qw(EINTR EWOULDBLOCK);

sub new {
    my ($class, $socket, $out) = @_;
    $socket->blocking(0);
    my $connection = bless {
        handle	   => $socket,
        fileno	   => fileno($socket),
        reading	   => 1,
        in_buffer  => "",
        out_buffer => $out,
        peer_ip    => $socket->peerhost,
        peer_port  => $socket->peerport,
    }, $class;
    vec($rd_mask, $connection->{fileno}, 1) = 1;
    vec($wr_mask, $connection->{fileno}, 1) = 1 if $out ne "";
    print STDERR "Handling connection from $connection->{peer_ip}:$connection->{peer_port}\n";
    return $connection;
}

sub readable {
    my $connection = shift;

    my $rc = sysread($connection->{handle}, my $buffer, 8192);
    if ($rc) {
        $connection->{in_buffer} .= $buffer;
        return;
    }
    if (defined($rc)) {
        # EOF
        printf STDERR "Read %d bytes\n", length($connection->{in_buffer});
        $connection->{reading} = 0;
        vec($rd_mask, $connection->{fileno}, 1) = 0;
        shutdown($connection->{handle}, 0);
        delete $connections{$connection} if $connection->{out_buffer} eq "";
        return;
    }
    return if $! == EINTR || $! == EWOULDBLOCK;
    print STDERR "Error reading from socket to $connection->{peer_ip}:$connection->{peer_port}: $!\n";
    delete $connections{$connection};
    vec($rd_mask, $connection->{fileno}, 1) = 0;
    vec($wr_mask, $connection->{fileno}, 1) = 0;
    $connection->{reading} = 0;
    close $connection->{handle};
}

sub writable {
    my $connection = shift;

    if (defined(my $rc = syswrite($connection->{handle}, 
                                  $connection->{out_buffer}))) {
        substr($connection->{out_buffer}, 0, $rc, "");
        if ($connection->{out_buffer} eq "") {
            vec($wr_mask, $connection->{fileno}, 1) = 0;
            shutdown($connection->{handle}, 1);
            delete $connections{$connection} if !$connection->{reading};
        }
        return;
    }
    return if $! == EINTR || $! == EWOULDBLOCK;
    print STDERR "Error writing to socket to $connection->{peer_ip}:$connection->{peer_port}: $!\n";
    delete $connections{$connection};
    vec($rd_mask, $connection->{fileno}, 1) = 0;
    vec($wr_mask, $connection->{fileno}, 1) = 0;
    $connection->{reading} = 0;
    close $connection->{handle};
}

sub DESTROY {
    my $connection = shift;
    print STDERR "Connection to $connection->{peer_ip}:$connection->{peer_port} forgotten\n";
}

package main;

$SIG{PIPE} = "IGNORE";

defined(my $file = shift) || die "Usage: $0 file\n";
open(my $fh, "<", $file) || die "Could not open $file: $!";
my $out = do { local $/; <$fh> };

my $listener = IO::Socket::INET->new(LocalAddr => "127.0.0.1",
                                     LocalPort => 1234,
                                     Listen    => 5) ||
    die "Could not listen on 127.0.0.1:1234: $!\n";

$listener->blocking(0);
my $listener_fd = fileno($listener);
vec($rd_mask, $listener_fd, 1) = 1;

while (1) {
    my $rc = select(my $rd = $rd_mask, my $wr = $wr_mask, undef, undef);
    if ($rc < 0) {
        next if $! == EINTR;
        die "Unexpected select error: $!";
    }
    for my $connection (values %connections) {
        $connection->writable if vec($wr, $connection->{fileno}, 1);
        $connection->readable if vec($rd, $connection->{fileno}, 1) && $connection->{reading};
    }
    if (vec($rd, $listener_fd, 1)) {
        if (my $socket = $listener->accept) {
            my $connection = Connection->new($socket, $out);
            $connections{$connection} = $connection;
        } else {
            print STDERR "Error accepting connection: $!\n" unless $! == EINTR;
        }
    }
}
